#!/usr/bin/env python

# Copyright (c) 2008 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import os
import sys
import re
import types
import pprint
import logging

# Komodo modules.
import process

log = logging.getLogger('gitlib')
#log.setLevel(logging.DEBUG)

#---- exceptions

class GITLibError(Exception):
    pass

class GIT:
    """A proxy to the Git command line app."""
    def __init__(self, executable='git', **options):
        self.executable = executable
        self.optd = options
        # Abortable process handle.
        self._processHelper = process.AbortableProcessHelper()

    def _runCommand(self, argv, cwd=None, env=None, input=None,
                    ignore_exit_status=False, terminalHandler=None):
        """Prepare and run the given arg vector, 'argv', and return the
        results.  Returns (<stdout>, <stderr>).
        """
        if type(argv) not in (types.ListType, types.TupleType):
            raise GITLibError("_runCommand:: argv is not a list: %r" % (argv, ))
        # Remove any empty options from the cmd list.
        argv = [ x for x in argv if x ]
        log.debug("Running %r in cwd %r", argv, cwd)
        output = None
        if env is None:
            import koprocessutils
            env = koprocessutils.getUserEnv()

        # Komodo can only handle svn messages in english.
        # http://bugs.activestate.com/show_bug.cgi?id=45677
        env['LC_MESSAGES'] = 'en_US'
        # Set LANGUAGE too, otherwise it may still come back in another lang
        # http://bugs.activestate.com/show_bug.cgi?id=68615
        env['LANGUAGE'] = 'en_US'

        # A light wrapper around process.ProcessOpen.
        p = self._processHelper.ProcessOpen(cmd=argv, cwd=cwd, env=env,
                                            universal_newlines=True)
        try:
            if input:
                input = input.encode("utf-8")
            if terminalHandler:
                terminalHandler.hookIO(p.stdin, p.stdout, p.stderr, " ".join(argv))
                p.wait()
                # Output and errors have gone to the terminal.
                output = ''
                error = ''
            else:
                output, error = p.communicate(input)
            retval = p.returncode
            p.close()
        finally:
            self._processHelper.ProcessDone()

        if retval != 0 and not ignore_exit_status:
            raise GITLibError("Error %s running '%s': \n%s"\
                              % (retval, argv, ''.join(error)))

        return output, error

    def abort(self):
        self._processHelper.ProcessAbort()

    def add(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "add"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def clone(self, urls, cwd=None, altdirname=None, command_options="",
              terminalHandler=None, env=None):
        if type(urls) in types.StringTypes:
            urls = [urls]
        argv = [self.executable, 'clone']
        if command_options:
            # Example: "--rev 103 --clean"
            argv += command_options.split(" ")
        argv += urls
        if altdirname:
            argv.append(altdirname)
        output, error = self._runCommand(argv,
                                         cwd=cwd,
                                         env=env,
                                         terminalHandler=terminalHandler)
        result = {}
        result['stdout'] = output
        result['stderr'] = error
        return result

    def commit(self, files, cwd=None, options=None, env=None, msgpath=None):
        result = {}
        argv = [self.executable, "commit"]
        if options:
            argv += options.split(" ")
        if msgpath is not None:
            argv += [ "-F", msgpath ]
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def diff(self, files, cwd=None, options=None, external=False, env=None):
        result = {}
        if external:
            argv = [self.executable, "difftool"]
        else:
            argv = [self.executable, "diff"]
        if options:
            argv += options.split(" ")
        if files:
            argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    log_revisions_re = re.compile(
        r'''commit\s+(?P<revision>\w*?)\n''' \
         '''Author:\s+(?P<author>.*?)\n''' \
         '''Date:\s+(?P<date>.*?)\n''' \
         '''\n'''
    )

    def _formatLogMessage(self, message):
        result = []
        for line in message.rstrip().splitlines(0):
            if line.startswith("    "):
                result.append(line[4:])
            else:
                result.append(line)
        return "\n".join(result) + "\n"

    def log(self, path, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "log"]
        if options:
            argv += options.split(" ")
        argv.append(path)
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error

        log_items = []
        matchIterator = self.log_revisions_re.finditer(output)
        try:
            match = matchIterator.next()
        except StopIteration:
            match = None
        while match:
            text_start = match.end()
            try:
                matchNext = matchIterator.next()
            except StopIteration:
                matchNext = None
            if matchNext:
                message = output[text_start:matchNext.start()]
            else:
                message = output[text_start:]

            log_items.append( {
                'revision': match.group('revision'),
                'author': match.group('author'),
                'date': match.group('date'),
                'message': self._formatLogMessage(message)
                } )
            match = matchNext

        result['log'] = log_items
        return result

    def pull(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "pull"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def remove(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "rm"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    def revert(self, files, cwd=None, options=None, env=None):
        result = {}
        argv = [self.executable, "checkout"]
        if options:
            argv += options.split(" ")
        argv += files
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error
        return result

    status_re = re.compile(r"^(?P<status>\w|\?)\s+(?P<path>.*)$")
    def status(self, filepaths=None, cwd=None, options=None, env=None,
               recursive=False):
        result = {}
        argv = [self.executable, "ls-files", "-t"]
        if options is None:
            # Default options:
            #  -m : modified
            #  -d : deleted
            #  -c : cached (repo files)
            #  -o : other files (non repo files)
            #  -k : conflicts
            options = "-m -d -o -c -k"
        argv += options.split(" ")
        if filepaths:
            argv += filepaths
        output, error = self._runCommand(argv, cwd=cwd, env=env)
        result['stdout'] = output
        result['stderr'] = error

        files = {}
        for line in output.splitlines(0):
            match = self.status_re.match(line)
            if match:
                hit = match.groupdict()
                path = hit['path']
                if (not recursive and "/" in path):
                    continue
                fullpath = os.path.join(cwd, path)
                files[fullpath] = { 'status': hit['status'],
                                    'relpath': path }
        result['files'] = files

        # Unfortunately, ls-files by itself does not give us enough information,
        # it does not distinguish between added files and modified files. We
        # use a second call to the status command to obtain the extra details.

        argv = [self.executable, "status"]
        if filepaths:
            argv += filepaths
        # Notes from the git-status manual page:
            # If there is no path that is different between the index file and
            # the current HEAD commit (i.e., there is nothing to commit by
            # running git commit), the command exits with non-zero status.
        # ARGH!!!!
        output, error = self._runCommand(argv, cwd=cwd, env=env,
                                         ignore_exit_status=True)
        if error:
            result['stderr'] += error

        section = ""
        for line in output.splitlines(0):
            if re.match(r"^# \w+", line):
                if line.startswith("# Changes to be committed:"):
                    section = "to_be_committed"
                elif line.startswith("# Changed but not updated:"):
                    section = "to_be_updated"
                else:
                    section = ""
            elif section:
                if section == "to_be_committed":
                    if "new file:" in line:
                        m = re.match(r"#\s+new file:\s+(.*)$", line)
                        if m:
                            path = m.group(1)
                            if (not recursive and "/" in path) or path.startswith(".."):
                                # This file resides in another directory.
                                continue
                            fullpath = os.path.join(cwd, path)
                            files[fullpath] = { 'status': 'A',
                                                'relpath': path }
                    elif "deleted:" in line:
                        m = re.match(r"#\s+deleted:\s+(.*)$", line)
                        if m:
                            path = m.group(1)
                            if (not recursive and "/" in path) or path.startswith(".."):
                                # This file resides in another directory.
                                continue
                            fullpath = os.path.join(cwd, path)
                            files[fullpath] = { 'status': 'D',
                                                'relpath': path }
                elif section == "to_be_updated":
                    m = re.match(r"#\s+modified:\s+(.*)$", line)
                    if m:
                        path = m.group(1)
                        if (not recursive and "/" in path) or path.startswith(".."):
                            # This file resides in another directory.
                            continue
                        fullpath = os.path.join(cwd, path)
                        files[fullpath] = { 'status': 'C',
                                            'relpath': path }

        return result
